/*
 Initially taken from Bakalab <https://github.com/bakalaborg/bakalab>
 Modified by Vít Skalický 2020
*/
package cz.vitskalicky.lepsirozvrh;

import android.content.ActivityNotFoundException;
import android.content.ClipData;
import android.content.ClipboardManager;
import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.os.Build;
import android.os.Handler;
import android.os.Looper;
import android.view.View;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.appcompat.app.AlertDialog;

import androidx.preference.PreferenceManager;
import com.google.android.material.snackbar.Snackbar;

import cz.vitskalicky.lepsirozvrh.model.RozvrhRecord;
import cz.vitskalicky.lepsirozvrh.notification.PermanentNotification;
import org.joda.time.DateTimeConstants;
import org.joda.time.LocalDate;
import org.joda.time.LocalTime;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Scanner;

import cz.vitskalicky.lepsirozvrh.model.rozvrh.Rozvrh;

public class Utils {
    public static final String TAG = Utils.class.getSimpleName();

    public static LocalDate getWeekMonday(LocalDate date) {
        if (date == null) return null;
        if (date == Rozvrh.Companion.getPERM()) return date;
        return date.dayOfWeek().setCopy(DateTimeConstants.MONDAY);

    }

    public static String dateToString(LocalDate date) {
        DateTimeFormatter dtf = DateTimeFormat.forPattern("yyyyMMdd");
        return dtf.print(date);
    }

    public static LocalDate parseDate(String date) {
        if (date == null || date.equals("")) return null;
        DateTimeFormatter dtf = DateTimeFormat.forPattern("yyyyMMdd");
        return dtf.parseLocalDate(date);
    }

    public static LocalDate getCurrentMonday() {
        return getWeekMonday(LocalDate.now());
    }

    public static LocalDate getDisplayWeekMonday(Context context) {
        int offset = KotlinUtils.INSTANCE.getWeekSwitchOffset(context);

        return getWeekMonday(LocalDate.now().plusDays(offset));
    }

    public static @Nullable RozvrhRecord.Key getNotificationRozvrhKey(@NonNull Context context){
        long accountid = PreferenceManager
                .getDefaultSharedPreferences(context)
                .getLong(PrefsConsts.NOTIFICATION_ACCOUNT, PermanentNotification.ACCOUNT_NOTIFICATION_LOGGED_OUT);
        if (!PermanentNotification.INSTANCE.isNotificationAccountValid(accountid)){
            return null;
        }
        return new RozvrhRecord.Key(SharedPrefs.getLong(context, PrefsConsts.NOTIFICATION_ACCOUNT), getCurrentMonday());
    }

    /**
     * For debugging purposes
     */
    public static String getDebugTime() {
        DateTimeFormatter dtf = DateTimeFormat.forPattern("mm:ss.SSS");
        LocalTime time = LocalTime.now();
        return dtf.print(time);
    }

    /**
     * Get fucking localized string for week info
     *
     * @param week week relative to now: 0 - current, 1 - next, -1 previous, {@code Integer.MAX_VALUE} permanent schedule
     * @return Localize, human friendly string
     */
    @SuppressWarnings("ConstantConditions")
    public static String getfl10nedWeekString(int week, Context context) {
        switch (week) {
            case 0:
                return context.getString(R.string.info_this_week);
            case 1:
                return context.getString(R.string.info_next_week);
            case -1:
                return context.getString(R.string.info_last_week);
            case Integer.MAX_VALUE:
                return context.getString(R.string.info_permanent);
        }
        if (week > 0)
            return context.getResources().getQuantityString(R.plurals.info_weeks_forward, week, week);
        else
            return context.getResources().getQuantityString(R.plurals.info_weeks_back, -1 * week, -1 * week);
    }

    public static interface Listener{
        public void method();
    }

    /**
     * Asks the user to send his Rozvrh because it is weird.

     * @param forToast View to use for displaying snackbar
     * @param rozvrhDate rozvrh date, null for permanent
     */
    public static void wtfRozvrh(Context context, View forToast, LocalDate rozvrhDate){

        //similar to {@link cz.vitskalicky.lepsirozvrh.settings.SettingsFragment#sendFeedback(boolean, Context, View)}

        LocalDate silent;
        try {
            silent = Utils.parseDate(SharedPrefs.getString(context, SharedPrefs.DISABLE_WTF_ROZVRH_UP_TO_DATE));
        }catch (IllegalArgumentException e){
            silent = null;
        }

        if (silent != null && LocalDate.now().isBefore(silent)){
            return;
        }

        AlertDialog ad = new AlertDialog.Builder(context)
                .setTitle(R.string.wtf_rozvrh_title)
                .setMessage(R.string.wtf_rozvrh_message)
                .setPositiveButton(R.string.wtf_rozvrh_report, (dialog, which) -> {
                    String body = null;
                    try {
                        body = context.getPackageManager().getPackageInfo(context.getPackageName(), 0).versionName;
                        body = "\n\n-----------------------------\n" + context.getString(R.string.email_message) + "\n Device OS: Android \n Device OS version: " +
                                Build.VERSION.RELEASE + "\n App Version: " + body + "\n Commit hash: " + BuildConfig.GitHash + "Build type: " + BuildConfig.BUILD_TYPE + "\n Device Brand: " + Build.BRAND +
                                "\n Device Model: " + Build.MODEL + "\n Device Manufacturer: " + Build.MANUFACTURER;
                        String sentryId = SharedPrefs.getString(context, SharedPrefs.SENTRY_ID);
                        if (sentryId != null && !sentryId.isEmpty()){
                            body += "\n Sentry client id: " + sentryId;
                        }else {
                            body += "\n Sentry client id not available";
                        }
                        body += "\n Sentry enabled: " + (new SharedPrefsKt(context)).getBoolean(PrefsConsts.ENABLE_SENTRY);
                        final String finBody = body;
                        new Thread(() -> {
                            String fileName;
                            if (rozvrhDate != null) {
                                fileName = "rozvrh-" + Utils.dateToString(Utils.getWeekMonday(rozvrhDate)) + ".xml";
                            } else {
                                fileName = "rozvrh-perm.xml";
                            }

                            String rozvrh = "";
                            try (FileInputStream inputStream = context.openFileInput(fileName)) {
                                //converts inputStream to string
                                Scanner s = new Scanner(inputStream).useDelimiter("\\A");
                                rozvrh = s.hasNext() ? s.next() : "";
                            } catch (FileNotFoundException e) {
                                rozvrh = "File not found: " + e.getMessage();
                            } catch (IOException e) {
                                rozvrh = "IOException: " + e.getMessage();
                            }

                            String finRozvrh = rozvrh;

                            new Handler(Looper.getMainLooper()).post(() -> {
                                String newBody = finBody + "\nSchedule:\n\n" + finRozvrh + "\n";

                                Intent intent = new Intent(Intent.ACTION_SEND);
                                intent.setType("message/rfc822");
                                String address = context.getString(R.string.CONTACT_MAIL);
                                intent.putExtra(Intent.EXTRA_EMAIL, new String[]{address});
                                intent.putExtra(Intent.EXTRA_SUBJECT, "");
                                intent.putExtra(Intent.EXTRA_TEXT, newBody);

                                try {
                                    context.startActivity(Intent.createChooser(intent, context.getString(R.string.send_email)));
                                } catch (ActivityNotFoundException ex) {
                                    Snackbar snackbar = Snackbar.make(forToast, context.getText(R.string.no_email_client), Snackbar.LENGTH_LONG);
                                    snackbar.setAction(R.string.copy_address, v -> {
                                        ClipboardManager clipboard = (ClipboardManager) context.getSystemService(Context.CLIPBOARD_SERVICE);
                                        ClipData clip = ClipData.newPlainText(address + "\n\n" + newBody, address + "\n\n" + newBody);
                                        clipboard.setPrimaryClip(clip);
                                        Snackbar.make(forToast, R.string.copied_to_clipboard, Snackbar.LENGTH_SHORT).show();
                                    });
                                    snackbar.show();
                                }
                            });

                        }).start();
                    } catch (PackageManager.NameNotFoundException e) {
                        Toast.makeText(context,"!",Toast.LENGTH_SHORT).show();
                    }
                })
                .setNegativeButton(R.string.wtf_rozvrh_later, (dialog, which) -> {})
                .setNeutralButton(R.string.wtf_not_fot_month, (dialog, which) -> {
                    SharedPrefs.setString(context, SharedPrefs.DISABLE_WTF_ROZVRH_UP_TO_DATE, Utils.dateToString(LocalDate.now().plusDays(30)));
                })
                .create();
        ad.show();
    }
}
